/*************************************************************************
 * Meso-H
 * two wire temperature transmitter
 *
 * Copyright (C) INOR Instrumentation AB 1997
 *
 * the Meso is a two wire temperature transmitter that supports the
 * Universal Commands and Common Practice Commands 34, 35, 40-42
 * 48, 49 and 59.  In addition the Meso-H has device specific commands
 * 130-141.
 */

/*************************************************************************
 *Revision History
 *
 * 01/04 31August1998 Niclas Palm
 * this revision was made to correct some syntax errors that ocurred
 * when the dd was tokenized with a new tokenizer belonging to the next
 * generation HART communicator (275) operating system (4.x)
 *
 * 01/03 2June1997 Wally Pratt
 * Corrected the loop current label on the root_menu.  added CLASS 
 * construct to the method missing one (loop_test, edit_err_offsets)
 *
 * 01/02 27May1997 Wally Pratt
 * made corrections as per the request from Niclas Palm via E-Mail 22-23
 * May 1997.  this includes more consistant naming of the USL and LSL.
 *
 * 01/02 19May1997 Wally Pratt
 * this is the revision that was produced while at INOR in Sweden.  All of
 * the Meso functionality is implemented and tested.  Addtitional comments
 * have been added as part of the cleanup.  however no functionality or
 * user visible strings have been changed since departing Sweden.
 * 
 * 01/01 17Apr1997 Wally Pratt
 * Initial Revision.
 * The initial revision supports all Meso-H capabilities except the
 * automatic calculation of custom linearization polynomials (commands
 * 138-141). This revision uses string literals only.  Standard
 * dictionary strings will be added in the next revision.
 *
 */

/* the standard DD include files */

#include "bi_codes.h"
#include "macros.ddl"

/* 
 * Declare manufacturer and device type
 */
MANUFACTURER INOR, DEVICE_TYPE _MESO, DEVICE_REVISION 1, DD_REVISION 4

/*************************************************************************
 * defines
 * this section contains the macros used in the DD
 *
 *************************************************************************
 */

#define DEBUG 0

#if DEBUG
/* the units supported by the meso depend on the sensor selected */
#define MESO_UNITS	\
			UNITS_CODE (32),				/* degrees C */		\
			UNITS_CODE (33),				/* degrees F */		\
			UNITS_CODE (37),				/* Ohm */			\
			UNITS_CODE (36)					/* mV */			\

#else
#define MESO_UNITS	\
	IF ( (sensor_type >= 128 && sensor_type <= 138) ||			\
		(sensor_type <= 17)	) {									\
			UNITS_CODE (32),				/* degrees C */		\
			UNITS_CODE (33)					/* degrees F */		\
	}															\
	ELSE {														\
		IF (sensor_type == 32) {								\
			UNITS_CODE (37)					/* Ohm */			\
		}														\
		ELSE {													\
			UNITS_CODE (36)					/* mV */			\
		}														\
	}
#endif

/*************************************************************************
 * IMPORTS
 * several standard DDs are always imported by a device DD.  these
 * are support the universal and common practice commands.  we are
 * importing the universal and common practice commands DDs, the common
 * tables as well as the MENU standard DD
 *
 *************************************************************************
 */
  
/*************************************************************************
 * Common Tables
 */

IMPORT STANDARD _TABLES, DEVICE_REVISION 9, DD_REVISION 2
{
	EVERYTHING;
}

/*************************************************************************
 * Universal Commmands
 */

IMPORT STANDARD _UNIVERSAL, DEVICE_REVISION 5, DD_REVISION 1
{
	EVERYTHING;

	REDEFINITIONS
	{
		/* this will need to be redefined for every device */
		VARIABLE device_type
		{
			REDEFINE TYPE ENUMERATED
			{
				{ 0xef,	"Meso" }
			}
		}

		/* define INOR as the manufacturer and distributor */
		VARIABLE manufacturer_id {
			REDEFINE TYPE ENUMERATED {
/* Begin CPK */	{ 0x5b,	"INOR" }
			}
		}
		
		VARIABLE private_label_distributor {
			REDEFINE TYPE ENUMERATED {
/* Begin CPK */ { 0x5b,	"INOR" }
			}
		}
		
		/* the transaction will need to be redefine to show the correct */
		/* number of variables returned in this command */
		COMMAND read_dynamic_variables_and_pv_current
		{
			REDEFINE TRANSACTION
			{
				REQUEST
				{}

				REPLY
				{
					response_code, device_status,
					INOR_loop_current, 
					PV.DIGITAL_UNITS, PV.DIGITAL_VALUE,
					SV.DIGITAL_UNITS, SV.DIGITAL_VALUE
				}
			}
		}
	}	/* end redefinitions */
}	/* end universal command import */

/*************************************************************************
 * Common Practice Commands
 */

IMPORT STANDARD _COMMON_PRACTICE, DEVICE_REVISION 7, DD_REVISION 2
{

	COMMAND write_pv_damping_value;					/* command 34 */
	COMMAND write_pv_range_values;					/* command 35 */
	COMMAND reset_configuration_change_flag;		/* command 38 */
	COMMAND enter_exit_fixed_pv_current_mode;		/* command 40 */
	COMMAND perform_transmitter_self_test;			/* command 41 */
	COMMAND perform_master_reset;					/* command 42 */
	
	VARIABLE xmtr_specific_status_0;
	COMMAND read_additional_transmitter_status;		/* command 48 */
	
	COMMAND write_pv_sensor_serial_number;			/* command 49 */

	VARIABLE response_preambles;
	COMMAND write_number_of_response_preambles;		/* command 59 */

	METHOD loop_test;
	METHOD return_to_normal;
	METHOD leave_fixed_current_mode;

	REDEFINITIONS
	{
		/* we had to redefine command 35 because the meso uses it
		 * to change the PV units.  Yes, I know this doesn't meet the
		 * HART Specs- This will be corrected next revision when
		 * command 44 is added
		 */
		COMMAND write_pv_range_values {
			REDEFINE TRANSACTION {
				REQUEST {
					PV.RANGE_UNITS,
					PV.UPPER_RANGE_VALUE, PV.LOWER_RANGE_VALUE
				}

				REPLY {
					response_code, device_status,
					PV.RANGE_UNITS, 
					PV.UPPER_RANGE_VALUE, PV.LOWER_RANGE_VALUE
				}
			}
		}

		/* only one byte of command 48 is meaningful. and of it only
		 * the lower two bits.  the other bits are tagged as 
		 * IGNORE_IN_HANDHELD because the meso sometimes returns an
		 * 0xFF (apparently on accident)
		 */
		VARIABLE xmtr_specific_status_0 {
			REDEFINE TYPE BIT_ENUMERATED {
				{ 0x01,	"EEPROM Failue",	HARDWARE },
				{ 0x02,	"ADC Error",		HARDWARE },
				{ 0x04, "",					MISC & IGNORE_IN_HANDHELD},
				{ 0x08, "",					MISC & IGNORE_IN_HANDHELD},
				{ 0x10, "",					MISC & IGNORE_IN_HANDHELD},
				{ 0x20, "",					MISC & IGNORE_IN_HANDHELD},
				{ 0x40, "",					MISC & IGNORE_IN_HANDHELD},
				{ 0x80, "",					MISC & IGNORE_IN_HANDHELD}
			}
		}

		COMMAND read_additional_transmitter_status {
			REDEFINE TRANSACTION {
				REQUEST {}

				REPLY {
					response_code, device_status,
					xmtr_specific_status_0
				}
			}
		}
		/* must redefine the standard loop test method because the meso
		 * cannot take 20.0 mA setting.  it does take 19.999 though
		 */
		REDEFINE METHOD loop_test {
			LABEL [loop_test];
			HELP [loop_test_help];
			CLASS ANALOG_OUTPUT;
			DEFINITION {
				char status[STATUS_SIZE];
				int choice;
				long var_id[1];
				int use_abort_method;
				char disp_string1[60];
				char disp_string2[60];
				int slen1;
				int slen2;

				slen1=60;
				slen2=60;
				get_dictionary_string(remove_loop_from_auto,disp_string1,slen1);
				ACKNOWLEDGE(disp_string1);
				add_abort_method(return_to_normal);

				IGNORE_ALL_RESPONSE_CODES();
				IGNORE_ALL_DEVICE_STATUS();
				XMTR_IGNORE_ALL_DEVICE_STATUS();
				XMTR_IGNORE_ALL_RESPONSE_CODES();

				use_abort_method = 0;
				do {
					get_dictionary_string(choose_analog_output_level,disp_string1,slen1);
					get_dictionary_string(choices_4_20_other_end,disp_string2,slen2);
					choice = SELECT_FROM_LIST(disp_string1, disp_string2);
					switch ( choice ) {
					case 0:
						assign_float(INOR_loop_current,4.0);
						break;

					case 1:
						assign_float(INOR_loop_current, 19.99999);
						break;

					case 2:
						get_dictionary_string(loop_output,disp_string1,slen1);
						if (fvar_value( INOR_loop_current ) > 19.0) {
							assign_float(INOR_loop_current, 19.0);
						}
						GET_DEV_VAR_VALUE(disp_string1, INOR_loop_current);
						break;

					case 3:
					default:
						process_abort();
						break;
					}
					send(40,status);
					if (!use_abort_method) {
						if (( status[STATUS_RESPONSE_CODE] != 0)
								&& ( status[STATUS_RESPONSE_CODE] != 3)
								&& ( status[STATUS_RESPONSE_CODE] != 4)) {
							display_response_status(40,status[STATUS_RESPONSE_CODE]);
							get_dictionary_string(abort_last_error,disp_string1,slen1);
							DELAY(2,disp_string1);
							process_abort();
						}
						if ( status[STATUS_RESPONSE_CODE] == 0) {
							remove_abort_method(return_to_normal);
							add_abort_method(leave_fixed_current_mode);
							use_abort_method =1;
						}
					}
					if ( status[STATUS_RESPONSE_CODE] ) {
						display_response_status(40,status[STATUS_RESPONSE_CODE]);
						if (( status[STATUS_RESPONSE_CODE] != 3)
								&& ( status[STATUS_RESPONSE_CODE] != 4)) {
							get_dictionary_string(abort_last_error,disp_string1,slen1);
							DELAY(2,disp_string1);
							process_abort();
						}
					}
					else {
						var_id[0] = VARID(INOR_loop_current);
						get_dictionary_string(fld_dev_otpt_fixed,disp_string1,slen1);
						acknowledge(disp_string1,var_id);
					}
				} while ( choice < 3);
				process_abort();
			}
		}
	} /* end redefinitions */
} /* end common pracice import */



/*************************************************************************
 * NOTE: MENU imported at end of file with other menu's
 * -----
 */



/*************************************************************************
 * VARIABLES
 * this begins the variables defininitions.  the first set of definitions
 * are in about the same order as the pv collection
 *
 *************************************************************************
 */

/*************************************************************************
 * the basic variables used in making the primary measurement
 * temperature value, units, USL, LSL & min_span
 */
VARIABLE temperature_value
{
	LABEL [pv];
	HELP [digital_value_help];
	CLASS CORRECTION & DYNAMIC;
	HANDLING READ;
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".1f";
	}
}

VARIABLE temperature_units
{
	LABEL [pv_sensor_unit];
	HELP [digital_units_help];
	CLASS CORRECTION;
	HANDLING READ;
	TYPE ENUMERATED
	{
		/* this is the macro from beginning of the file */
		MESO_UNITS
	}
}

VARIABLE temperature_LSL
{
	LABEL [pv_lsl];
	HELP [upper_sensor_limit_help];
	CLASS CORRECTION;
	HANDLING READ;
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".1f";
	}
}

VARIABLE temperature_USL
{
	LABEL [pv_usl];
	HELP [upper_sensor_limit_help];
	CLASS CORRECTION;
	HANDLING READ;
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".1f";
	}
}

VARIABLE temperature_min_span
{
	LABEL [pv_minimum_span];
	HELP [minimum_span_help];
	CLASS CORRECTION;
	HANDLING READ;
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".1f";
	}
}

VARIABLE temperature_damping
{
	LABEL [pv_damping_value];
	HELP [seconds_damping_value_help];
	CLASS CORRECTION;
	CONSTANT_UNIT [sec];
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".1f";
		EDIT_FORMAT "5.1f";
	}
}

/*************************************************************************
 * the variables used to convert the measurement into percent range
 */
VARIABLE input_percent_range
{
	LABEL [pv_percent_range];
	HELP [percent_of_range_help];
	CLASS INPUT & DYNAMIC;
	CONSTANT_UNIT [Percent];
	HANDLING READ;
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".2f";
	}
}

VARIABLE xfer_function
{
	LABEL [transfer_function];
	HELP [transfer_function_help];
	CLASS INPUT;
	HANDLING READ;
	TYPE ENUMERATED
	{
		TRANSFER_FUNCTION_CODE (0),
		TRANSFER_FUNCTION_CODE (250),
		TRANSFER_FUNCTION_CODE (251),
		TRANSFER_FUNCTION_CODE (252),
		TRANSFER_FUNCTION_CODE (253)
	}
}

/*************************************************************************
 * the URV, LRV and units variables
 */
VARIABLE input_units
{
	HELP [range_units_help];
	CLASS INPUT;
	LABEL [pv_range_units];

	TYPE ENUMERATED {
		MESO_UNITS
	}
}

VARIABLE temperature_URV
{
	HELP [upper_range_value_help];
	LABEL [pv_urv];
	CLASS INPUT;
	TYPE FLOAT {
		DISPLAY_FORMAT ".1f";
		EDIT_FORMAT "5.1f";
	}
}

VARIABLE temperature_LRV {
	HELP [lower_range_value_help];
	LABEL [pv_lrv];
	CLASS INPUT;
	TYPE FLOAT {
		DISPLAY_FORMAT ".1f";
		EDIT_FORMAT "5.1f";
	}
}

/*************************************************************************
 * Loop Current variables
 */
VARIABLE INOR_loop_current
{
	LABEL "Loop mA";
	HELP [analog_value_help];
	CLASS DEVICE & DYNAMIC;
	CONSTANT_UNIT [mA];
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".2f";
		EDIT_FORMAT "5.3f";
	}
}

/* the analog upper and lower endpoints are rather unique.  HART assumes
 * the end point for the primary variable are always 4 and 20 mA.  The
 * meso allows the upper and lower endpoints to be moved! the urv and lrv
 * are still defined at those end points.  I don't know if this is an
 * interoperability problem or not.
 */
VARIABLE analog_lower_endpt {
	LABEL	"AO@PV LRV";
	HELP	[analog_output_lower_end_point_help];
	CLASS ANALOG_OUTPUT;
	CONSTANT_UNIT [mA];
	TYPE FLOAT {
		DISPLAY_FORMAT	".1f";
		EDIT_FORMAT		"5.3f";

		MIN_VALUE	 4.0;
		MAX_VALUE	20.0;
	}
}
VARIABLE	analog_upper_endpt {
	LABEL	"AO@PV URV";
	HELP	[analog_output_upper_end_point_help];
	CLASS ANALOG_OUTPUT;
	CONSTANT_UNIT [mA];
	TYPE FLOAT {
		DISPLAY_FORMAT	".1f";
		EDIT_FORMAT		"5.3f";

		MIN_VALUE	 4.0;
		MAX_VALUE	20.0;
	}
}

VARIABLE analog_alarm_code
{
	LABEL [alarm_select];
	HELP [alarm_code_help];
	CLASS ANALOG_OUTPUT;
	HANDLING READ;
	TYPE ENUMERATED
	{
		ALARM_SELECTION_CODE(0),
		ALARM_SELECTION_CODE(1),
		ALARM_SELECTION_CODE(239),
		ALARM_SELECTION_CODE(250),
		ALARM_SELECTION_CODE(251),
		ALARM_SELECTION_CODE(252),
		ALARM_SELECTION_CODE(253)
	}
}

/*************************************************************************
 * serial number of the probe connected
 */
VARIABLE probe_serial_no {
	HELP [sensor_serial_number_help];
	CLASS DEVICE;
	LABEL [pv_sensor_serial_number];
	HANDLING READ;
	TYPE UNSIGNED_INTEGER (3) {
		DISPLAY_FORMAT	"7d";
	}
}

/* Here is where the real fun begins! since different sensors have differing
 * wiring styles and other characteristics almost everything in the meso
 * keys off the sensor type and the wiring to the sensor.
 */
VARIABLE sensor_type {
	LABEL	"Type";
	HELP	"Sensor Type - Indicates what kind of sensor that is connected to the field device";
	CLASS DEVICE;
	TYPE ENUMERATED {
		{ 1,	[pt100_a_385] },
		{ 2,	"PT1000, a=385" },
		{ 3,	"D100, a=392" },

		{ 4,	"PTx, a=385" },

		{ 16,	[ni100] },
		{ 17,	"Ni1000" },

		{ 32,	[ohms] },
		{ 64,	[millivolts] },

		{ 128,	[t_c_type_b] },

		{ 129,	[t_c_type_e] },
		{ 130,	[t_c_type_j] },
		{ 131,	[t_c_type_k] },
		{ 132,	[t_c_type_din_l] },
		{ 133,	[t_c_type_n] },
		{ 134,	[t_c_type_r] },
		{ 135,	[t_c_type_s] },
		{ 136,	[t_c_type_t] },
		{ 137,	[t_c_type_din_u] },
		{ 138,	"T/C typ A/E" },

		{ 160,	"T/C typ X" }
	}
}

VARIABLE sensor_connection {
	LABEL "Wiring";
	HELP "Connection - Indicates how the sensor is connected to the field device and if the field device monitors the sensor isolation";
	CLASS CORRECTION;
	TYPE ENUMERATED
	{
#if DEBUG
				{ 5,	"Diff." },
				{ 16,	"SmartSense" },
				{ 2,	[two_wires] },
				{ 3,	[three_wires] },
				{ 4,	[four_wires] }
#else
	/* the sensor type determines the wiring that can be used */
		IF ( (sensor_type >= 128 && sensor_type <= 138) ||
			(sensor_type == 160) ){
				{ 16,	"SmartSense" },		/* thermocouples are SmartSense or */
				{ 2,	[two_wires] }		/* 2 wire only */
		} 
		ELSE {
			IF (sensor_type == 1) {			/* PT100 can do smart sense */
				{ 5,	"2 Wire Diff." },
				{ 16,	"SmartSense" },
				{ 3,	[three_wires] },
				{ 4,	[four_wires] }
			}
			ELSE {
				IF (sensor_type == 64) {	/* mV always 2 wire */
					{ 2,	[two_wires] }
				}
				ELSE {
					{ 3,	[three_wires] },/* rtd and ohm do 3 or 4 wire */
					{ 4,	[four_wires] }
				}
			}
		}
#endif
	} /* end enumeration */
}

VARIABLE sensor_CJC {
	LABEL "SnsrCJC";
	HELP "Cold Junction Compensation - Indicates whether the field device performs cold junction compensation";

	CLASS CORRECTION;
	TYPE ENUMERATED 
	{
#if DEBUG 
				{ 0, [on] },
				{ 1, [off] }
#else
		/* most T/C use CJC */
		IF (sensor_type > 128) {
				{ 0, [on] },
				{ 1, [off] }
		} 
		ELSE {	/* Type B may not use CJC */
			IF ( sensor_type == 128 ) {
				{ 1, [off] }
			}
			ELSE { /* everyone else don't use CJC (eg, RTD's) */
				{ 1, [off] }
			}
		}
#endif
	}
}

VARIABLE sensor_linearization {
	LABEL "LinType";
	HELP "Linearization - Indicates the type of linearization to be performed";


	CLASS CORRECTION;
	TYPE ENUMERATED {
#if DEBUG 
		{ 0, "Tmpture" }
		{ 1, "Linear" },
		{ 2, "Custom" }
#else
		IF (sensor_type == 160 || sensor_type == 32 ||
					sensor_type == 64) {
			{ 1, "Linear" },	/* TC-X, ohm, and mV can be linear or use */
			{ 2, "Custom" }		/* a custom linearization polynomial */
		}
		ELSE {
			{ 0, "Tmpture" }	/* most T/C and RTDs can only use temperature linearization */
		}
#endif
	}
}

VARIABLE ptX_resistance {	/* only used for custom platinum RTD's */
	LABEL "PT-X Ohms";
	HELP "PtX Resistance - Indicates the sensor resistance at 0 degrees C / 32 degrees F";

	CLASS CORRECTION;
	TYPE FLOAT {
		MIN_VALUE 10.0;
		MAX_VALUE 1000.0;
	}
	CONSTANT_UNIT "Ohm";
}

/* this variable determines the isolation resistance that will trip the
 * SmartSense alarm.  if there is less resistance measured between the 
 * shield and the sensor then something must be wrong
 */
VARIABLE ssense_kOhm {
	LABEL "SmartSns";
	HELP "Isolation Resistance - Indicates the resistance limit for the low isolation detection";

	CLASS CORRECTION;
	TYPE FLOAT {
		MIN_VALUE 50.0;
		MAX_VALUE 200.0;
	}
	CONSTANT_UNIT "kOhm";
}

/*************************************************************************
 *the secondary measurement related variable declarations
 */
VARIABLE temp_comp_value
{
	LABEL "DevTemp";
	HELP "Device Temperature - Transmitter internal temperature.  Used for cold junction compensation.";
	CLASS CORRECTION & DYNAMIC;
	HANDLING READ;
	TYPE FLOAT
	{
		DISPLAY_FORMAT ".1f";
	}
}

VARIABLE temp_comp_units
{
	LABEL "DevTempUnit";
	HELP [digital_units_help];
	CLASS CORRECTION;
	HANDLING READ;
	TYPE ENUMERATED
	{
		UNITS_CODE (32)	/* degrees C */
	}
}

/*************************************************************************
 * the variables used to set the alarm levels.  there are two possible
 * alarm levels.  the first is the alarm level associated with a 
 * sensor break.  this alarm level works with all sensor types EXCEPT
 * millivolt.  this alarm is disabled by writing a zero as the set point
 *
 * the second alarm is smart sense.  the meso monitors the resistance
 * between shield and the sensing element.  smart sense works with
 * thermocouples or pt100 sensors only
 */
VARIABLE sensor_break_error_setpoint {
	LABEL "SnsrBreak";
	HELP "Sensor Break - Indicates which level the analog output should go to when a sensor break occurs.  Set to 0.0mA to disable sensor break alarm.";
	CLASS ANALOG_OUTPUT;
	CONSTANT_UNIT [mA];
	TYPE FLOAT {
		DISPLAY_FORMAT	".1f";
		EDIT_FORMAT		".3f";

		MIN_VALUE	 0.0;
		MAX_VALUE	22.8;
	}
}

VARIABLE sensor_isolation_error_setpoint {
	HELP "Smart Sense - Indicates which level the analog output should go to when the sensor isolation is lower than the defined isolation resistance";
	LABEL "SnsrIsolat";
	CLASS ANALOG_OUTPUT;
	CONSTANT_UNIT [mA];
	TYPE FLOAT {
		DISPLAY_FORMAT	".1f";
		EDIT_FORMAT		".3f";

		MIN_VALUE	 3.6;
		MAX_VALUE	22.8;
	}
}

/*************************************************************************
 * the variables used for sensor trim.  the approach for trimming the
 * sensor by the meso involves entering the sensor error at the urv
 * and lrv.  the sensor error entries have the display and edit 
 * precision modified based on the magnitude of the urv and lrv
 */
VARIABLE sensor_error_units {
	CLASS CORRECTION;
	LABEL "Error Units";
	HELP "Error Units - Indicates which unit to be displayed with the sensor error at the URV and the sensor error at the LRV";
	TYPE ENUMERATED {
		MESO_UNITS
	}
}

VARIABLE sensor_error_URV {
	CLASS CORRECTION;
	LABEL "Err@URV";
	HELP "Sensor Error at the URV - Defines the sensor deviation from the correct value at the URV";
	TYPE FLOAT {
		DISPLAY_FORMAT IF ((temperature_URV >= 1000.0) || 
			(temperature_LRV >= 1000.0)) { 
				".1f";
			}
			ELSE {
				IF ((temperature_URV >= 100.0) || 
					(temperature_LRV >= 100.0)) { 
						".2f";
				}
				ELSE {
						".3g";
				}
			}

		EDIT_FORMAT IF ((temperature_URV >= 1000.0) ||
			(temperature_LRV >= 1000.0)) { 
				"5.2f";
			}
			ELSE {
				IF ((temperature_URV >= 100.0) || 
					(temperature_LRV >= 100.0)) { 
						"5.3f";
				}
				ELSE {
						"5.4g";
				}
			}
	}
}

VARIABLE sensor_error_LRV {
	CLASS CORRECTION;
	LABEL "Err@LRV";
	HELP "Sensor error at the LRV - Defines the sensor deviation from the correct value at the LRV";
	TYPE FLOAT {
		DISPLAY_FORMAT IF ((temperature_URV >= 1000.0) || 
			(temperature_LRV >= 1000.0)) { 
				".1f";
			}
			ELSE {
				IF ((temperature_URV >= 100.0) || 
					(temperature_LRV >= 100.0)) { 
						".2f";
				}
				ELSE {
						".3g";
				}
			}

		EDIT_FORMAT IF ((temperature_URV >= 1000.0) ||
			(temperature_LRV >= 1000.0)) { 
				"5.2f";
			}
			ELSE {
				IF ((temperature_URV >= 100.0) || 
					(temperature_LRV >= 100.0)) { 
						"5.3f";
				}
				ELSE {
						"5.4g";
				}
			}
	}
}

/*************************************************************************
 * the custom engineering calculation. another interesting feature allows
 * the entry and storing of special engineering ranges and units.  The
 * meso doesn't actually use the values for anything.  however,  they are
 * only available with the ohm, mV and TCx sensors using custom 
 * linearization.  the idea is that after the custom linearization the
 * mA correspond to the engineering range.  software (and this DD) can
 * read this range and calculate the engineering value from the mA.
 */
VARIABLE engineering_val_unit_code {
	CLASS CORRECTION;
	LABEL "Eng Units";
	HELP "Engineering unit - Defines which unit to be displayed with the engineering value at the URV and the engineering value at the LRV";
	TYPE ENUMERATED {
	/* Temperature */
		UNITS_CODE (32),		/* degC */
		UNITS_CODE (33),		/* degF */

	/* Pressure */
		UNITS_CODE (7),			/* bar */
		UNITS_CODE (8),			/* mbar */
		UNITS_CODE (11),		/* PA */
		UNITS_CODE (12),		/* kPA */
		UNITS_CODE (237),		/* mega_pascals */

	/* Electromagnetic unit of Electric Potential */
		UNITS_CODE (36),		/* mV */
		UNITS_CODE (58),		/* v */

	/* Electrostatic unit of Current */
		UNITS_CODE (39),		/* mA */

	/* Time */
		UNITS_CODE (51)			/* sec */

	/* Miscellaneous */
		UNITS_CODE (57),		/* Percent */
		UNITS_CODE (59),		/* pH */
		UNITS_CODE (139),		/* parts_per_million */

	/* Special Units */
		{ 240, "us" },
		{ 241, "ms" },
		{ 242, "/min" },
		{ 243, "-" }
	}
}

VARIABLE engineering_val_URV {
	CLASS CORRECTION;
	LABEL "Eng URV";
	HELP "Engineering value at the URV - Defines the engineering value associated with the PV unit at the URV";
	TYPE FLOAT {
		DISPLAY_FORMAT ".1f";
		EDIT_FORMAT		".3f";
	}
}

VARIABLE engineering_val_LRV {
	CLASS CORRECTION;
	LABEL "Eng LRV";
	HELP "Engineering value at the LRV - Defines the engineering value associated with the PV unit at the LRV";
	TYPE FLOAT {
		DISPLAY_FORMAT ".1f";
		EDIT_FORMAT		".3f";
	}
}

/*************************************************************************
 * the custom linearization coefficients.  these are only available when
 * an ohm, mV or TCx sensor is used.  this allows the meso to support
 * things like strain gage measurements.  since this is a cubic
 * polynomial and there are 4 terms that may be entered.  someday
 * the DD may calculate these directly.
 */

#if DEBUG
#define	CUSTOM_LINEAR TRUE;
#else
#define CUSTOM_LINEAR IF ( sensor_type != 32 && sensor_type != 64 && \
		sensor_type != 160 || sensor_linearization != 2) \
			{ FALSE; }
#endif

VARIABLE lin_offset_coef_A0 {
	CLASS CORRECTION;
	LABEL "Offset(A0)";
	HELP "Offset (A0) - Defines the constant part of the customized linearization polynomial";
	TYPE FLOAT {
				DISPLAY_FORMAT	".3g";
				EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_LINEAR
}

VARIABLE lin_line_coef_A1 {
	CLASS CORRECTION;
	LABEL "Line  (A1)";
	HELP "Line (A1) - Defines the linear part of the customized linearization polynomial";
	TYPE FLOAT {
				DISPLAY_FORMAT	".3g";
				EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_LINEAR
}

VARIABLE lin_square_coef_A2 {
	CLASS CORRECTION;
	LABEL "Square(A2)";
	HELP "Square (A2) - Defines the square part of the customized linearization polynomial";
	TYPE FLOAT {
				DISPLAY_FORMAT	".3g";
				EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_LINEAR
}

VARIABLE lin_cube_coef_A3 {
	CLASS CORRECTION;
	LABEL "Cube  (A3)";
	HELP "Cube (A3) - Defines the cubic part of the customized linearization polynomial";
	TYPE FLOAT {
				DISPLAY_FORMAT	".3g";
				EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_LINEAR
}

/*************************************************************************
 *the custom cold junction compensation coefficients
 */
#if DEBUG
#define	CUSTOM_CJC TRUE;
#else
#define	CUSTOM_CJC IF ( sensor_type != 160 || sensor_CJC != 0 ) \
		{ FALSE; }
#endif

VARIABLE cjc_line_coef_B1 {
	CLASS CORRECTION;
	LABEL "Line  (B1)";
	HELP "Line (B1) - Defines the linear part of the customized cold junction compensation";
	TYPE FLOAT {
		DISPLAY_FORMAT	".3g";
		EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_CJC
}

VARIABLE cjc_square_coef_B2 {
	CLASS CORRECTION;
	LABEL "Square(B2)";
	HELP "Square (B2) - Defines the square part of the customized cold junction compensation";
	TYPE FLOAT {
		DISPLAY_FORMAT	".3g";
		EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_CJC
}

VARIABLE cjc_cube_coef_B3 {
	CLASS CORRECTION;
	LABEL "Cube  (B3)";
	HELP "Cube (B3) - Defines the cubic part of the customized cold junction compensation";
	TYPE FLOAT {
		DISPLAY_FORMAT	".3g";
		EDIT_FORMAT		".3g";
	}
	VALIDITY CUSTOM_CJC
}

/*************************************************************************
 * Relationships
 * declares the relations used by a host to tag stale data in the real
 * time data base for updating.
 *
 *************************************************************************
 */

UNIT pv_units_relation {	/* the variables affected by the pv units */
	temperature_units:
		temperature_value,
		temperature_USL,
		temperature_LSL,
		temperature_min_span
}

UNIT sv_units_relation {	/* the variables affected by the sv units */
	temp_comp_units:
		temp_comp_value
}

UNIT input_units_relation {	/* changing the input units affects the URV & LRV values */
	input_units:
		temperature_URV,
		temperature_LRV
}

UNIT sensor_error_units_relation {
	sensor_error_units:
		sensor_error_URV, 
		sensor_error_LRV
}

UNIT eng_val_units_relation {
	engineering_val_unit_code:
		engineering_val_URV, 
		engineering_val_LRV
}

REFRESH pv_unit_refresh {	/* zzzz wap check does changing the units change URV & LRV */
	temperature_units:
		input_units,
		temperature_URV,
		temperature_LRV
}

WRITE_AS_ONE ao_range_wao {
	analog_lower_endpt,
	analog_upper_endpt
}

REFRESH sensor_refresh {
	sensor_type:
		temperature_URV, temperature_LRV, input_units
		temperature_units, temperature_USL, temperature_LSL, temperature_min_span
}

/*************************************************************************
 * ARRAYS and COLLECTIONS
 * this section declares the arrays and collections.  these are generally
 * used by the universal and common practice command standard DDs
 *
 *************************************************************************
 */

/*************************************************************************
 * Dynamic Variables Configuration
 */
ARRAY OF COLLECTION dynamic_variables {
	ELEMENTS {
		PRIMARY, pv, [pv];
		SECONDARY, sv, [sv];
	}
}

COLLECTION OF VARIABLE pv {
	LABEL [pv];
	HELP [primary_variable_help];
	MEMBERS {
/* Sensor Correction Information */
		DIGITAL_UNITS,			temperature_units,	[units],			[digital_units_help];
		DIGITAL_VALUE,			temperature_value,	[blank],			[digital_value_help];
		SENSOR_UNITS,			temperature_units,	[units],			[digital_units_help];
		UPPER_SENSOR_LIMIT,		temperature_USL,	[usl],				[upper_sensor_limit_help];
		LOWER_SENSOR_LIMIT,		temperature_LSL,	[lsl],				[lower_sensor_limit_help];
		MINIMUM_SPAN,			temperature_min_span,[minimum_span],	[minimum_span_help];
		DAMPING_VALUE,			temperature_damping,[damping_value],	[seconds_damping_value_help];

/* Range Information */
		PERCENT_RANGE,			input_percent_range,[percent_of_range],	[percent_of_range_help];
		TRANSFER_FUNCTION,		xfer_function,		[transfer_function],[transfer_function_help];
		RANGE_UNITS,			input_units,		[range_units],		[range_units_help];
		UPPER_RANGE_VALUE,		temperature_URV,	[urv],				[upper_range_value_help];
		LOWER_RANGE_VALUE,		temperature_LRV,	[lrv],				[lower_range_value_help];

/* Loop Current Information */
		ANALOG_VALUE,			INOR_loop_current,	"Loop mA",			[analog_value_help];
		ALARM_CODE,				analog_alarm_code,	[alarm_select],		[alarm_code_help];
		LOWER_ENDPOINT_VALUE,	analog_lower_endpt,	[analog_output_lower_endpoint],	[analog_output_lower_end_point_help];
		UPPER_ENDPOINT_VALUE,	analog_upper_endpt,	[analog_output_upper_endpoint],	[analog_output_upper_end_point_help];

/* Device Information */
		SENSOR_SERIAL_NUMBER,	probe_serial_no,	[sensor_serial_number], [sensor_serial_number_help];
	}
}

COLLECTION OF VARIABLE sv {
	LABEL [sv];
	HELP [secondary_variable_help];
	MEMBERS {
		DIGITAL_UNITS,			temp_comp_units,	[units],			[digital_units_help];
		DIGITAL_VALUE,			temp_comp_value,	[blank],			[digital_value_help];

	}
}

COLLECTION OF VARIABLE SENSOR
{
	LABEL [sensor];
	HELP [sensor_correction_block_help];
	MEMBERS
	{
		DIGITAL_VALUE,		temperature_value,		[blank],			[digital_value_help];
		SENSOR_UNITS,		temperature_units,		[units],			[digital_units_help];
		UPPER_SENSOR_LIMIT,	temperature_USL,		[usl],				[upper_sensor_limit_help];
		LOWER_SENSOR_LIMIT,	temperature_LSL,		[lsl],				[lower_sensor_limit_help];
		MINIMUM_SPAN,		temperature_min_span,	[minimum_span],		[minimum_span_help];
		DAMPING_VALUE,		temperature_damping,	[damping_value],	[seconds_damping_value_help];
	}
}

COLLECTION OF VARIABLE INPT
{
	LABEL [analog_input];
	HELP [analog_input_block_help];
	MEMBERS
	{
		PERCENT_RANGE,		input_percent_range,	[percent_of_range],	[percent_of_range_help];
		TRANSFER_FUNCTION,	xfer_function,			[transfer_function],[transfer_function_help];
		RANGE_UNITS,		input_units,			[range_units],		[range_units_help];
		UPPER_RANGE_VALUE,	temperature_URV,		[urv],				[upper_range_value_help];
		LOWER_RANGE_VALUE,	temperature_LRV,		[lrv],				[lower_range_value_help];
	}
}

COLLECTION OF VARIABLE ANALOG
{
	LABEL "Loop mA";
	HELP [analog_output_block_help];
	MEMBERS
	{
		ANALOG_VALUE,		INOR_loop_current,		[blank],			[analog_value_help];
		ALARM_CODE,			analog_alarm_code,		[alarm_type],		[alarm_code_help];
		UPPER_ENDPOINT_VALUE,	analog_upper_endpt,	[analog_output_upper_endpoint],	[analog_output_upper_end_point_help];
		LOWER_ENDPOINT_VALUE,	analog_lower_endpt,	[analog_output_lower_endpoint],	[analog_output_lower_end_point_help];

	}
}

COLLECTION OF VARIABLE DEV
{
	LABEL [device];
	HELP [device_block_help];
	MEMBERS
	{
		SENSOR_SERIAL_NUMBER,	probe_serial_no,	[sensor_serial_number], [sensor_serial_number_help];
	}
}

/*************************************************************************
 * Commands
 * Now to define the device specific commands. These include:
 *
 * 130	read_input_configuration
 * 131	write_input_configuration
 * 132	read_output_configuration
 * 133	write_output_configuration
 * 134	read_sensor_error_range_values
 * 135	write_sensor_error_range_values
 * 136	read_engineering_range_values
 * 137	write_engineering_range_values
 * 138	read_custom_linearization
 * 139	write_custom_linearization
 * 140	read_custom_cjc_linearization
 * 141	write_custom_cjc_linearization
 *
 *************************************************************************
 */

COMMAND read_input_configuration {
	NUMBER 130;
	OPERATION READ;
	TRANSACTION {
		REQUEST { }
		REPLY {
			response_code, device_status,
			sensor_type, sensor_connection, 
			sensor_linearization, sensor_CJC,
			ptX_resistance, ssense_kOhm
		}
	}

	RESPONSE_CODES {
		0,		SUCCESS,		[no_command_specific_errors];
	}
}

COMMAND write_input_configuration {
	NUMBER 131;
	OPERATION WRITE;
	TRANSACTION {
		REQUEST {
			sensor_type, sensor_connection, 
			sensor_linearization, sensor_CJC,
			ptX_resistance, ssense_kOhm
		}
		REPLY {
			response_code, device_status,
			sensor_type, sensor_connection, 
			sensor_linearization, sensor_CJC,
			ptX_resistance, ssense_kOhm
		}
	}

	RESPONSE_CODES {
		0,	SUCCESS,			[no_command_specific_errors];
		3,	DATA_ENTRY_ERROR,	[passed_parameter_too_large];
		4,	DATA_ENTRY_ERROR,	[passed_parameter_too_small];
		5,	MISC_ERROR,			[too_few_data_bytes_recieved];
		6,	MISC_ERROR,			[xmtr_specific_command_error];
		7,	MODE_ERROR,			[in_write_protect_mode];

		11,	DATA_ENTRY_ERROR,	[invalid_sensor_type_code];
		12,	DATA_ENTRY_ERROR,	"Invalid No. of Wires/ Connection";
	}
}

COMMAND read_output_configuration {
	NUMBER 132;
	OPERATION READ;
	TRANSACTION {
		REQUEST {
		}
		REPLY {
			response_code, device_status,
			sensor_break_error_setpoint,
			analog_upper_endpt,	analog_lower_endpt,
			sensor_isolation_error_setpoint
		}
	}

	RESPONSE_CODES {
		0,		SUCCESS,		[no_command_specific_errors];
	}
}

COMMAND write_output_configuration {
	NUMBER 133;
	OPERATION WRITE;
	TRANSACTION {
		REQUEST {
			sensor_break_error_setpoint,
			analog_upper_endpt,	analog_lower_endpt,
			sensor_isolation_error_setpoint
		}
		REPLY {
			response_code, device_status,
			sensor_break_error_setpoint,
			analog_upper_endpt,	analog_lower_endpt,
			sensor_isolation_error_setpoint
		}
	}

	RESPONSE_CODES {
		0,	SUCCESS,			[no_command_specific_errors];
		3,	DATA_ENTRY_ERROR,	[passed_parameter_too_large];
		4,	DATA_ENTRY_ERROR,	[passed_parameter_too_small];
		5,	MISC_ERROR,			[too_few_data_bytes_recieved];
		7,	MODE_ERROR,			[in_write_protect_mode];
		9,	DATA_ENTRY_ERROR,	"LRV AO Endpoint to Hi";
		10,	DATA_ENTRY_ERROR,	"LRV AO Endpoint to Lo";
		11,	DATA_ENTRY_ERROR,	"URV AO Endpoint to Hi";
		12,	DATA_ENTRY_ERROR,	"URV AO Endpoint to Lo";
		14,	DATA_ENTRY_ERROR,	[span_too_small];
	}
}

COMMAND read_sensor_error_range_values {
	NUMBER 134;
	OPERATION READ;
	TRANSACTION {
		REQUEST {
		}
		REPLY {
			response_code, device_status,
			sensor_error_units,
			sensor_error_URV, sensor_error_LRV
		}
	}

	RESPONSE_CODES {
		0,		SUCCESS,		[no_command_specific_errors];
	}
}

COMMAND write_sensor_error_range_values {
	NUMBER 135;
	OPERATION WRITE;
	TRANSACTION {
		REQUEST {
			sensor_error_units,
			sensor_error_URV, sensor_error_LRV
		}
		REPLY {
			response_code, device_status,
			sensor_error_units,
			sensor_error_URV, sensor_error_LRV
		}
	}

	RESPONSE_CODES {
		0,	SUCCESS,			[no_command_specific_errors];
		2,	DATA_ENTRY_ERROR,	[invalid_selection]; 
		5,	MISC_ERROR,			[too_few_data_bytes_recieved];
		7,	MODE_ERROR,			[in_write_protect_mode];

		9,	DATA_ENTRY_ERROR,	"LRV Sensor Error to Hi";
		10,	DATA_ENTRY_ERROR,	"LRV Sensor Error to Lo";
		11,	DATA_ENTRY_ERROR,	"URV Sensor Error to Hi";
		12,	DATA_ENTRY_ERROR,	"URV Sensor Error to Lo";

		14,	DATA_ENTRY_ERROR,	[span_too_small];
	}
}

COMMAND read_engineering_range_values {
	NUMBER 136;
	OPERATION READ;
	TRANSACTION {
		REQUEST {
		}
		REPLY {
			response_code, device_status,
			engineering_val_unit_code,
			engineering_val_URV, engineering_val_LRV
		}
	}

	RESPONSE_CODES {
		0,		SUCCESS,		[no_command_specific_errors];
	}
}

COMMAND write_engineering_range_values {
	NUMBER 137;
	OPERATION WRITE;
	TRANSACTION {
		REQUEST {
			engineering_val_unit_code,
			engineering_val_URV, engineering_val_LRV
		}
		REPLY {
			response_code, device_status,
			engineering_val_unit_code,
			engineering_val_URV, engineering_val_LRV
		}
	}

	RESPONSE_CODES {
		0,	SUCCESS,			[no_command_specific_errors];
	}
}

COMMAND read_custom_linearization {
	NUMBER 138;
	OPERATION READ;
	TRANSACTION {
		REQUEST {
		}
		REPLY {
			response_code, device_status,
			lin_offset_coef_A0,
			lin_line_coef_A1,
			lin_square_coef_A2,
			lin_cube_coef_A3
		}
	}

	RESPONSE_CODES {
		0,		SUCCESS,		[no_command_specific_errors];
	}
}

COMMAND write_custom_linearization {
	NUMBER 139;
	OPERATION WRITE;
	TRANSACTION {
		REQUEST {
			lin_offset_coef_A0,
			lin_line_coef_A1,
			lin_square_coef_A2,
			lin_cube_coef_A3
		}
		REPLY {
			response_code, device_status,
			lin_offset_coef_A0,
			lin_line_coef_A1,
			lin_square_coef_A2,
			lin_cube_coef_A3
		}
	}

	RESPONSE_CODES {
		0,	SUCCESS,			[no_command_specific_errors];
	}
}

COMMAND read_custom_cjc_linearization {
	NUMBER 140;
	OPERATION READ;
	TRANSACTION {
		REQUEST {
		}
		REPLY {
			response_code, device_status,
			cjc_line_coef_B1,
			cjc_square_coef_B2,
			cjc_cube_coef_B3
		}
	}

	RESPONSE_CODES {
		0,		SUCCESS,		[no_command_specific_errors];
	}
}

COMMAND write_custom_cjc_linearization {
	NUMBER 141;
	OPERATION WRITE;
	TRANSACTION {
		REQUEST {
			cjc_line_coef_B1,
			cjc_square_coef_B2,
			cjc_cube_coef_B3
		}
		REPLY {
			response_code, device_status,
			cjc_line_coef_B1,
			cjc_square_coef_B2,
			cjc_cube_coef_B3
		}
	}

	RESPONSE_CODES {
		0,	SUCCESS,			[no_command_specific_errors];
	}
}




/*************************************************************************
 * MENUES
 * below is list of menu declaritions
 *
 *************************************************************************
 */
/*************************************************************************
 * Importing the Standard MENU
 */

IMPORT STANDARD _MENU, DEVICE_REVISION 1, DD_REVISION 1
{
	MENU root_menu;
		MENU device_setup;						/* no change */
			MENU process_variables;
			MENU diag_service;
				MENU test_device;
					MENU status_display;
			MENU basic_setup;
			MENU detailed_setup;
				MENU measuring_elements;
				MENU signal_conditioning;
				MENU hart_output;
				MENU device_info;
					MENU device_revisions;		/* no change */
			MENU review;
	
	MENU hot_key;

	WRITE_AS_ONE range_wao;
	EDIT_DISPLAY rerange;

	REDEFINITIONS
	{
		MENU root_menu {
			REDEFINE ITEMS {
				device_setup,					 				/* menu */
				PV.DIGITAL_VALUE (DISPLAY_VALUE),	   			/* variable */
				INOR_loop_current (DISPLAY_VALUE, READ_ONLY),	/* variable */
				PV.LOWER_RANGE_VALUE (DISPLAY_VALUE),			/* variable */
				PV.UPPER_RANGE_VALUE (DISPLAY_VALUE)			/* variable */
			}
		}

		MENU process_variables {
			REDEFINE ITEMS {
				temperature_value(DISPLAY_VALUE, READ_ONLY),
				input_percent_range(DISPLAY_VALUE, READ_ONLY),
				INOR_loop_current(DISPLAY_VALUE, READ_ONLY),
				temp_comp_value(DISPLAY_VALUE, READ_ONLY),
				calc_eng_val							/* Method */
			}
		}

		MENU diag_service {
			REDEFINE ITEMS {
				test_device,				/* menu */
				loop_test,					/* common-method */
				sensor_menus[sensor_type]	/* array */
			}
		}

		MENU test_device {
			REDEFINE ITEMS {
				status_display(REVIEW),	/* menu */
				self_test,				/* method */
				master_reset			/* method */
			}
		}
		MENU status_display {
			REDEFINE ITEMS {
				xmtr_specific_status_0 (DISPLAY_VALUE,READ_ONLY)
			}
		}

		MENU basic_setup {
			REDEFINE ITEMS {
				tag (DISPLAY_VALUE),				/* variable */
				PV.DIGITAL_UNITS (DISPLAY_VALUE),	/* variable */
				PV.DAMPING_VALUE (DISPLAY_VALUE), 	/* variable */
				rerange								/* Edit Display */
			}
		}

		MENU detailed_setup {
			REDEFINE ITEMS {
				measuring_elements,				/* menu */
				signal_conditioning,			/* menu */
				ao_menus[sensor_type],			/* menu */
				hart_output,					/* menu */
				device_info						/* menu */
			}
		}

		MENU measuring_elements {
			REDEFINE ITEMS {
				temperature_value (DISPLAY_VALUE, READ_ONLY),
				sensor_info_menus[sensor_type],
				select_sensor					/* method */
			}
		}

		MENU signal_conditioning {
			REDEFINE ITEMS {
				SENSOR.DAMPING_VALUE (DISPLAY_VALUE),
				INPT.UPPER_RANGE_VALUE (DISPLAY_VALUE),
				INPT.LOWER_RANGE_VALUE (DISPLAY_VALUE),
				input_percent_range (DISPLAY_VALUE, READ_ONLY)
			}
		}

		MENU hart_output {
			REDEFINE LABEL "HART";
			REDEFINE ITEMS {
				polling_address (DISPLAY_VALUE),
				request_preambles (DISPLAY_VALUE)
			}
		}

		MENU device_info {
			REDEFINE ITEMS {
				tag (DISPLAY_VALUE),
				descriptor (DISPLAY_VALUE),
				manufacturer_id (DISPLAY_VALUE, READ_ONLY),
				device_type (DISPLAY_VALUE),
				device_id (DISPLAY_VALUE),
				private_label_distributor (DISPLAY_VALUE),
				date (DISPLAY_VALUE),
				write_protect (DISPLAY_VALUE),
				message (DISPLAY_VALUE),
				sensor_type (DISPLAY_VALUE, READ_ONLY),
				probe_serial_no (DISPLAY_VALUE),
				final_assembly_number (DISPLAY_VALUE),
				device_revisions,	/* menu */
				status_display		/* menu */
			}
		}

		MENU review	{
			REDEFINE ITEMS {
				device_type,
				private_label_distributor,
				tag, date, descriptor, message,
				final_assembly_number,

				probe_serial_no,
				sensor_type, sensor_connection, sensor_linearization,
				sensor_CJC,
				ptX_resistance, ssense_kOhm,

				input_units,
				temperature_damping,
				temperature_min_span,
				temperature_LSL, temperature_USL,
				temperature_LRV, temperature_URV,
				analog_upper_endpt, analog_lower_endpt,

				sensor_break_error_setpoint,
				sensor_isolation_error_setpoint,

				sensor_error_units, sensor_error_URV, sensor_error_LRV,

				engineering_val_unit_code, 
				engineering_val_URV, engineering_val_LRV,

				lin_offset_coef_A0, lin_line_coef_A1,
				lin_square_coef_A2, lin_cube_coef_A3,
				cjc_line_coef_B1, cjc_square_coef_B2, cjc_cube_coef_B3
				xfer_function,

				write_protect,
				manufacturer_id,
				device_id,
				tag,
				descriptor,
				message,
				date,

				universal_revision,
				transmitter_revision,
				software_revision,
				polling_address,
				request_preambles
			}
		}
			
		EDIT_DISPLAY rerange	{

			REDEFINE EDIT_ITEMS { 
				temperature_LRV,
				analog_lower_endpt,
				temperature_URV,
				analog_upper_endpt,	
				PV.RANGE_UNITS }
		}
	} /* end MENU redefinitions */
} /* end MENU import */

/*************************************************************************
 * CJC/ Linearization Menus
 * now we manipulate the menu structure to allow the optional support
 * for custom CJC and custom linearization.  this is a little tricky.
 * first three dummy variables are declared to use as indexes to a
 * couple of levels of indirection thru some arrays of menus.  basically
 * the combinations of sensor numbers, CJC (off/on), and linearization
 * type (temperature, linear, or custom) will determine which menu
 * will be used.
 *
 *************************************************************************
 */
ARRAY OF MENU sensor_menus {
	ELEMENTS {
		32,		no_cjc_select[sensor_linearization];	/* ohms */
		64,		no_cjc_select[sensor_linearization];	/* millivolts */
		160,	line_select[sensor_linearization];	/* T/C typ X */

		1,		cal1;			/* pt100_a_385 */
		2,		cal1;			/* PT1000, a=385 */
		3,		cal1;			/* D100, a=392 */
		4,		cal1;			/* PTx, a=385 */
		16,		cal1;			/* ni100 */
		17,		cal1;			/* Ni1000 */
		128,	cal1;			/* t_c_type_b */
		129,	cal1;			/* t_c_type_e */
		130,	cal1;			/* t_c_type_j */
		131,	cal1;			/* t_c_type_k */
		132,	cal1;			/* t_c_type_din_l */
		133,	cal1;			/* t_c_type_n */
		134,	cal1;			/* t_c_type_r */
		135,	cal1;			/* t_c_type_s */
		136,	cal1;			/* t_c_type_t */
		137,	cal1;			/* t_c_type_din_u */
		138,	cal1;			/* T/C typ A/E */
	}
}

/* accessed if there is no choice about custom CJC but we might
 * have a custom linearization.  sensor_linearization is used 
 * to direct the reference to the right menu
 */
ARRAY OF MENU no_cjc_select {
	ELEMENTS {
		0,	cal1;
		1,	cal1;
		2,	cal2;
	}
}

/* accessed if there we can have both custom CJC and linearization
 * this array determines whether there is custom linearization
 */
ARRAY OF MENU line_select {
	ELEMENTS {
		0,	wo_line_cjc_menus[sensor_CJC];
		1,	wo_line_cjc_menus[sensor_CJC];
		2,	w_line_cjc_menus[sensor_CJC];
	}
}

/* the next two arrays determine if custom CJC is used
 * wo_l... is accessed from the line_select if custom linearization
 * is not used.  otherwise w_li... is accessed.  the value of 
 * sensor_CJC is used to reference the right menu entry.
 */
ARRAY OF MENU wo_line_cjc_menus {
	ELEMENTS {
		1,	cal1;
		0,	cal3;
	}
}

ARRAY OF MENU w_line_cjc_menus {
	ELEMENTS {
		1,	cal2;
		0,	cal4;
	}
}

/* below are the final four menus the are possible.  using the
 * arrays of menus above one of the menus below is selected
 * depending on whether custom CJC or linearization is allowed
 */
MENU cal1 {		/*no cjc entry, no custom linearization */
	LABEL "Configuration";
	ITEMS {
		rerange,		/* Edit display */
		set_err_offsets
	}
}

MENU cal2 {		/* no cjc entry */
	LABEL "Configuration";
	ITEMS {
		rerange,			/* Edit display */
		set_err_offsets,
		custom_linear,		/* Edit Display */
		custom_eng_range	/* Menu */
	}
}

MENU cal3 {		/* no custom linearization */
	LABEL "Configuration";
	ITEMS {
		rerange,		/* Edit display */
		set_err_offsets,
		custom_cjc		/* Edit display */
	}
}

MENU cal4 {		/* cjc entry, custom linearization */
	LABEL "Configuration";
	ITEMS {
		rerange,			/* Edit display */
		set_err_offsets,
		custom_cjc,			/* Edit display */
		custom_linear,		/* Edit Display */
		custom_eng_range	/* Menu */
	}
}

MENU custom_eng_range {		/* displays the custom engineering range */
	LABEL "Custom Range";
	ITEMS {
		engineering_val_unit_code(DISPLAY_VALUE),
		engineering_val_URV(DISPLAY_VALUE), 
		engineering_val_LRV(DISPLAY_VALUE),
		calc_eng_val				/* Method */
	}
}

MENU set_err_offsets {		/* allows review of sensor error values */
	LABEL "Sensor Error";
	ITEMS {
		temperature_LRV (DISPLAY_VALUE, READ_ONLY),
		sensor_error_LRV (DISPLAY_VALUE, READ_ONLY),
		temperature_URV (DISPLAY_VALUE, READ_ONLY),
		sensor_error_URV (DISPLAY_VALUE, READ_ONLY),
		edit_err_offsets
	}
}

/*************************************************************************
 * Sensor Info Menus
 * now we manipulate the menu structure to based on the sensor type
 * for some sensors the value of the smartSense, ptC resistance and
 * other attributes are fixed.  as a result we will not display them
 * if they are fixed
 *
 *************************************************************************
 */
ARRAY OF MENU sensor_info_menus {
	ELEMENTS {
		1,		sensor_info_pt100 [sensor_connection];	/* pt100_a_385 */
		2,		menu_sensor_pt;							/* PT1000, a=385 */
		3,		menu_sensor_pt;							/* D100, a=392 */
		4,		menu_sensor_ptx;						/* PTx, a=385 */
		16,		menu_sensor_pt;							/* ni100 */
		17,		menu_sensor_pt;							/* Ni1000 */

		32,		menu_sensor_ohm_mv;						/* ohms */
		64,		menu_sensor_ohm_mv;						/* millivolts */

		128,	sensor_info_tc128 [sensor_connection];	/* t_c_type_b */
		129,	sensor_info_tc [sensor_connection];		/* t_c_type_e */
		130,	sensor_info_tc [sensor_connection];		/* t_c_type_j */
		131,	sensor_info_tc [sensor_connection];		/* t_c_type_k */
		132,	sensor_info_tc [sensor_connection];		/* t_c_type_din_l */
		133,	sensor_info_tc [sensor_connection];		/* t_c_type_n */
		134,	sensor_info_tc [sensor_connection];		/* t_c_type_r */
		135,	sensor_info_tc [sensor_connection];		/* t_c_type_s */
		136,	sensor_info_tc [sensor_connection];		/* t_c_type_t */
		137,	sensor_info_tc [sensor_connection];		/* t_c_type_din_u */
		138,	sensor_info_tc [sensor_connection];		/* T/C typ A/E */

		160,	sensor_info_tcx [sensor_connection];	/* T/C typ X */
	}
}

/*************************************************************************
 * Normal TC Menus
 * now we manipulate the menu structure to based on the wiring type
 * for some thermocouples wiring can be smartSense or 2 wire.  if
 * SmartSense then we have a menu entry that allows them to modify the
 * SmartSense resistance
 *
 *************************************************************************
 */
ARRAY OF MENU sensor_info_tc {
	ELEMENTS {
		2,	menu_sensor_tc;
		3,	menu_sensor_tc;		/* should never get to this one!!!! */
		4, 	menu_sensor_tc;		/* should never get to this one!!!! */
		5, 	menu_sensor_tc;		/* should never get to this one!!!! */
		16,	menu_sensor_tc_smart;
	}
}

MENU menu_sensor_tc {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_CJC (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

MENU menu_sensor_tc_smart {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_CJC (DISPLAY_VALUE),
		ssense_kOhm (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

/*************************************************************************
 * TC type B Menus
 * now we manipulate the menu structure for type B T/C based on the wiring
 * type.  The wiring can be smartSense or 2 wire.  if SmartSense then we 
 * have a menu entry that allows them to modify the SmartSense resistance
 *
 *************************************************************************
 */
ARRAY OF MENU sensor_info_tc128 {
	ELEMENTS {
		2,	menu_sensor_tc128;
		3,	menu_sensor_tc128;		/* should never get to this one!!!! */
		4, 	menu_sensor_tc128;		/* should never get to this one!!!! */
		5, 	menu_sensor_tc128;		/* should never get to this one!!!! */
		16,	menu_sensor_tc128_smart;
	}
}

MENU menu_sensor_tc128 {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_menus[sensor_type]	/* array */
	}
}

MENU menu_sensor_tc128_smart {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		ssense_kOhm (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

/*************************************************************************
 * TC type X Menus
 * now we manipulate the menu structure for type X T/C based on the wiring
 * type.  The wiring can be smartSense or 2 wire.  if SmartSense then we 
 * have a menu entry that allows them to modify the SmartSense resistance
 *
 *************************************************************************
 */
ARRAY OF MENU sensor_info_tcx {
	ELEMENTS {
		2,	menu_sensor_tcx;
		3,	menu_sensor_tcx;		/* should never get to this one!!!! */
		4, 	menu_sensor_tcx;		/* should never get to this one!!!! */
		5, 	menu_sensor_tcx;		/* should never get to this one!!!! */
		16,	menu_sensor_tcx_smart;
	}
}

MENU menu_sensor_tcx {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_CJC (DISPLAY_VALUE),
		sensor_linearization (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

MENU menu_sensor_tcx_smart {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_CJC (DISPLAY_VALUE),
		ssense_kOhm (DISPLAY_VALUE),
		sensor_linearization (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

/*************************************************************************
 * Ohm/ mV Menu
 * now we manipulate the menu structure for type X T/C based on the wiring
 * type.  The wiring can be smartSense or 2 wire.  if SmartSense then we 
 * have a menu entry that allows them to modify the SmartSense resistance
 *
 *************************************************************************
 */
MENU menu_sensor_ohm_mv {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_linearization (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

/*************************************************************************
 * PT/Ni (rtd) Menus
 * the Pt100 is unique in that it can have SmartSense enabled.  as a 
 * result, now we manipulate the menu structure to based on the wiring type
 * if SmartSense then we have a menu entry that allows them to modify the
 * SmartSense resistance
 *
 *************************************************************************
 */
ARRAY OF MENU sensor_info_pt100 {
	ELEMENTS {
		2,	menu_sensor_pt;		/* should never get to this one!!!! */
		3,	menu_sensor_pt;
		4, 	menu_sensor_pt;
		5, 	menu_sensor_pt;
		16,	menu_sensor_pt_smart;
	}
}

MENU menu_sensor_pt {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		sensor_menus[sensor_type]	/* array */
	}
}

MENU menu_sensor_pt_smart {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		ssense_kOhm (DISPLAY_VALUE),
		sensor_menus[sensor_type]	/* array */
	}
}

MENU menu_sensor_ptx {
	LABEL [sensor_info];
	ITEMS {
		sensor_type (DISPLAY_VALUE, READ_ONLY),
		sensor_connection (DISPLAY_VALUE, READ_ONLY),
		ptX_resistance (DISPLAY_VALUE, READ_ONLY),
		sensor_menus[sensor_type]	/* array */
	}
}

/*************************************************************************
 * Analog Output Menus
 * now we manipulate the menu structure to allow the optional support
 * for the sensor break and smart sense alarm levels.  whether these two
 * alarm level settings are used depends on the sensor type and whether
 * smart sense wiring type is selected.
 *
 *************************************************************************
 */
ARRAY OF MENU ao_menus {
	ELEMENTS {
		1,		ao_connect[sensor_connection];	/* pt100_a_385 */
		2,		ao_sbreak;						/* PT1000, a=385 */
		3,		ao_sbreak;						/* D100, a=392 */
		4,		ao_sbreak;						/* PTx, a=385 */
		16,		ao_sbreak;						/* ni100 */
		17,		ao_sbreak;						/* Ni1000 */

		32,		ao_sbreak;						/* ohms */
		64,		ao_no_break;					/* millivolts */

		128,	ao_connect[sensor_connection];	/* t_c_type_b */
		129,	ao_connect[sensor_connection];	/* t_c_type_e */
		130,	ao_connect[sensor_connection];	/* t_c_type_j */
		131,	ao_connect[sensor_connection];	/* t_c_type_k */
		132,	ao_connect[sensor_connection];	/* t_c_type_din_l */
		133,	ao_connect[sensor_connection];	/* t_c_type_n */
		134,	ao_connect[sensor_connection];	/* t_c_type_r */
		135,	ao_connect[sensor_connection];	/* t_c_type_s */
		136,	ao_connect[sensor_connection];	/* t_c_type_t */
		137,	ao_connect[sensor_connection];	/* t_c_type_din_u */
		138,	ao_connect[sensor_connection];	/* T/C typ A/E */

		160,	ao_connect[sensor_connection];	/* T/C typ X */
	}
}

ARRAY OF MENU ao_connect {		/* used for indirection on types */
	ELEMENTS {					/* that are smart sense capable */
		2,	ao_sbreak;
		3,	ao_sbreak;
		4, 	ao_sbreak;
		5, 	ao_sbreak;
		16,	ao_smart;
	}
}

MENU ao_sbreak {		/* sensor break detection capable */
	LABEL [analog_output];
	ITEMS {	/* smart sense isolation detection */
		loop_test,
		end_points,
		INOR_loop_current (DISPLAY_VALUE,READ_ONLY),
		sensor_break_error_setpoint(DISPLAY_VALUE)
	}
}

MENU ao_smart {		/* sensor break detection capable & smart sense capable */
	LABEL [analog_output];
	ITEMS {	/* no isolation detection */
		loop_test,
		end_points,
		INOR_loop_current (DISPLAY_VALUE,READ_ONLY),
		sensor_break_error_setpoint(DISPLAY_VALUE),
		sensor_isolation_error_setpoint (DISPLAY_VALUE)
	}
}

MENU ao_no_break {		/* mV only: no sensor break detection, */
	LABEL [analog_output];
	ITEMS {	/* no isolation detection */
		loop_test,
		end_points,
		INOR_loop_current (DISPLAY_VALUE,READ_ONLY)
	}
}

MENU end_points {		/* allows access to the analog output end points */
	LABEL "AO End Points";
	ITEMS {
		analog_lower_endpt(DISPLAY_VALUE),
		analog_upper_endpt(DISPLAY_VALUE)	
	}
}

/*************************************************************************
 * used by the HC-275
 *
 * First this determines the variable write order. Second this is the list
 * of variables that the user will be allowed to view and modify under the
 * offline configure mode.
 */
MENU upload_variables
{
	LABEL "";
	ITEMS {
		input_units, temperature_damping,
		temperature_LRV, temperature_URV,
		analog_upper_endpt, analog_lower_endpt,

#if !DEBUG
		sensor_type, sensor_connection, sensor_linearization,
		sensor_CJC, probe_serial_no, ptX_resistance, ssense_kOhm,

		sensor_break_error_setpoint, sensor_isolation_error_setpoint,

		sensor_error_units, sensor_error_URV, sensor_error_LRV,

		engineering_val_unit_code, 
		engineering_val_URV, engineering_val_LRV,

		lin_offset_coef_A0, lin_line_coef_A1,
		lin_square_coef_A2, lin_cube_coef_A3,
		cjc_line_coef_B1, cjc_square_coef_B2, cjc_cube_coef_B3
		xfer_function,

		tag, date, descriptor, message,
#endif
		final_assembly_number
	}
}

/*************************************************************************
 * used by the HC-275
 *
 * tells the HC-275 that a variable could cause a change
 * in theloop current when it is modified.  this causes 
 * the HC-275 to issue a warning to the operator when that
 * variable is edited
 */
ARRAY OF VARIABLE loop_warning_variables
{
	ELEMENTS {
		1,	temperature_LRV;
		2,	temperature_URV;
		3,	input_units;
		4,	analog_upper_endpt;
		5,	analog_lower_endpt;

		6,	sensor_type;
		8,	sensor_linearization;
		9,	sensor_CJC;
		10,	polling_address;

#if !DEBUG
		11,	ptX_resistance;
		12,	ssense_kOhm;

		15,	sensor_error_units;
		16,	sensor_error_URV;
		17,	sensor_error_LRV;

		21,	lin_offset_coef_A0;
		22,	lin_line_coef_A1;
		23,	lin_square_coef_A2;
		24,	lin_cube_coef_A3;
		25,	cjc_line_coef_B1;
		26,	cjc_square_coef_B2;
		27,	cjc_cube_coef_B3;
#endif
	}
}

/*************************************************************************
 * EDIT_DISPLAYS
 * The following are the list of Edit Displays used by the Meso
 *
 *************************************************************************
 */
EDIT_DISPLAY custom_linear {
	LABEL "Enter Lin Coef";
	EDIT_ITEMS {
		lin_offset_coef_A0,
		lin_line_coef_A1,
		lin_square_coef_A2,
		lin_cube_coef_A3
	}
}

EDIT_DISPLAY custom_cjc {
	LABEL "Enter CJC Coef";
	EDIT_ITEMS {
		cjc_line_coef_B1,
		cjc_square_coef_B2,
		cjc_cube_coef_B3
	}
}

EDIT_DISPLAY kbd_sensor_trim {
	LABEL "Enter Sensor Error";
	DISPLAY_ITEMS {
		temperature_LRV,
		temperature_URV
	}
	EDIT_ITEMS {
		sensor_error_units,
		sensor_error_LRV,
		sensor_error_URV
	}
}


/*************************************************************************
 * METHODS
 *
 *************************************************************************
 */

/*************************************************************************
 * calc_eng_val
 *
 *************************************************************************
 */

METHOD calc_eng_val {
	LABEL "Calc Eng Value";
	HELP "Calculated engineering value - Indicates the PV scaled to the entered engineering unit";
	CLASS DEVICE;
	VALIDITY IF (sensor_linearization == 2) {TRUE;} ELSE {FALSE;}
	DEFINITION {
		char status[STATUS_SIZE];
		char disp_string[60];
		float y, span;
		long ids[6];
		int len;

		len = 60;

		/* update all the variable needed to caclulate the
		 * engineering value
		 */

				send_command(2);
		send_command(136);
#if 0
		ids[1] = VARID (engineering_val_URV);
		ids[2] = VARID (engineering_val_LRV);
		ids[3] = VARID (analog_upper_endpt);
		ids[4] = VARID (analog_lower_endpt);
		ids[5] = VARID (INOR_loop_current);
		acknowledge( "Eng URV = %{1}\nEng LRV = %{2}\nUpper EndPt = %{3}\nLower EndPt = %{4}", ids);
#endif
		span = (fvar_value (engineering_val_URV) - 
					fvar_value (engineering_val_LRV)) /
				(fvar_value (analog_upper_endpt) - 
					fvar_value (analog_lower_endpt));
		
		y = (fvar_value (INOR_loop_current) -
				fvar_value (analog_lower_endpt)) * span
				+ fvar_value (engineering_val_LRV);

		ids[0] = VARID (engineering_val_unit_code);
				acknowledge( "Engineering Value \n	%{y} %{0}", ids );
	}
}


#define UPDATE_IN_PROGRESS		8
#define DEVICE_BUSY				32
#define MAXIMUM_RESET_RETRIES	10

/*************************************************************************
 * self_test
 *
 *************************************************************************
 */

METHOD self_test
{
	LABEL	"Self Test" ;
	HELP "Self Test - Performs a internal self test in the field device that checks the device vital functions";
	CLASS DEVICE;
	DEFINITION
	{
		char status[STATUS_SIZE];
		char xmtr_status[MAX_XMTR_STATUS_LEN];
		char disp_string[60];
		long x;
		int len;
		int number_of_tries;

		len = 60;
		get_dictionary_string(remove_loop_from_auto, disp_string, len); 
		ACKNOWLEDGE(disp_string);
		add_abort_method(return_to_normal);

		get_dictionary_string(about_to_perform_limited_selftest,disp_string,len);
		ACKNOWLEDGE(disp_string);

		get_dictionary_string(performing_selftest,disp_string,len);
		PUT_MESSAGE(disp_string);

		send(41,status);

		if(status[STATUS_RESPONSE_CODE]){
			display_response_status(41,status[STATUS_RESPONSE_CODE]);
			get_dictionary_string(abort_last_error, disp_string, len);
			DELAY(2,disp_string);
			process_abort();
		}

		RETRY_ON_NO_DEVICE();
		number_of_tries = 0;

		do {
			get_more_status(status,xmtr_status);
		} while (((status[STATUS_RESPONSE_CODE] == DEVICE_BUSY)
				|| (status[STATUS_RESPONSE_CODE] == UPDATE_IN_PROGRESS))
			&& (number_of_tries++ < MAXIMUM_RESET_RETRIES));

		if ((status[STATUS_RESPONSE_CODE] == DEVICE_BUSY)
				|| (status[STATUS_RESPONSE_CODE] == UPDATE_IN_PROGRESS)) {
			ACKNOWLEDGE("Device maybe stuck in Self Test.");
		}
		else if (status[STATUS_RESPONSE_CODE]) {
			display_response_status(48,status[STATUS_RESPONSE_CODE]);
			get_dictionary_string(abort_last_error, disp_string, len);
						DELAY(2,disp_string);
		}
		else if ( xmtr_status[0] ) {

		/* check the errors returned by the device
		 * meso sometimes returns 0xff and we only know the 
		 * first two bits 
		 */
			for ( x=1; x < 0x04; x <<= 1) { 
				if ( xmtr_status[0] & x )
					display_xmtr_status(xmtr_specific_status_0, x);
			}
		}
		else {
			ACKNOWLEDGE("Self Test OK");
		}
		process_abort();		/* this will call return_to_normal */
	}
}

/*************************************************************************
 * master_reset
 *
 *************************************************************************
 */
METHOD master_reset
{
	LABEL	"Master Reset" ;
	HELP "Master Reset - Performs a master reset of the field device. Equivalent to cycling the transmitter power supply";
	CLASS DEVICE;			
	DEFINITION
	{
		char status[STATUS_SIZE];
		char xmtr_status[MAX_XMTR_STATUS_LEN];
		char disp_string1[60];
		long x;
		int slen1;
		int number_of_tries;

		slen1 = 60;
		get_dictionary_string(remove_loop_from_auto, disp_string1, slen1); 
		ACKNOWLEDGE(disp_string1);
		add_abort_method(return_to_normal);

		ACKNOWLEDGE("About to Preform Master Reset");

		send(42,status);

		if(status[STATUS_RESPONSE_CODE]){
			display_response_status(42,status[STATUS_RESPONSE_CODE]);
			get_dictionary_string(abort_last_error, disp_string1, slen1);
			DELAY(2,disp_string1);
			process_abort();
		}

		DELAY(8,"Master Reset in\nprogress please wait");

		SET_NUMBER_OF_RETRIES(7);
		RETRY_ON_NO_DEVICE();
		number_of_tries = 0;

		get_more_status(status,xmtr_status);
		get_more_status(status,xmtr_status);

		do {
			get_more_status(status,xmtr_status);
		} while (((status[STATUS_RESPONSE_CODE] == DEVICE_BUSY)
				|| (status[STATUS_RESPONSE_CODE] == UPDATE_IN_PROGRESS))
			&& (number_of_tries++ < MAXIMUM_RESET_RETRIES));

		if ((status[STATUS_RESPONSE_CODE] == DEVICE_BUSY)
				|| (status[STATUS_RESPONSE_CODE] == UPDATE_IN_PROGRESS)) {
			ACKNOWLEDGE("Device maybe stuck in Master Reset.");
		}
		else if (status[STATUS_RESPONSE_CODE]) {
			display_response_status(48,status[STATUS_RESPONSE_CODE]);
			get_dictionary_string(abort_last_error, disp_string1, slen1);
			DELAY(2,disp_string1);
		}
		else if ( xmtr_status[0] ) {

		/* check the errors returned by the device
		 * meso sometimes returns 0xff and we only know the 
		 * first two bits 
		 */
			for ( x=1; x < 0x04; x <<= 1) {
				if ( xmtr_status[0] & x )
					display_xmtr_status(xmtr_specific_status_0, x);
			}
		}
		else {
			ACKNOWLEDGE("Master reset OK");
		}
		process_abort();		/* this will call return_to_normal */
	}
}

/*************************************************************************
 * select_sensor
 *
 *************************************************************************
 */
METHOD select_sensor
{
	LABEL	"Select Sensor" ;
	HELP "Select Sensor - Select a sensor type and the different properties associated with this sensor type";
	CLASS CORRECTION;			
	DEFINITION
	{
		char status[STATUS_SIZE];	/* used for message processing */
		char xmtr_status[MAX_XMTR_STATUS_LEN];
		int err;

		long ids[5];				/* used to hold VARIDS */

		char disp_string[60];		/* used to get dictionary strings */
		int slen;
		
		int selection;		/* used in select from lists */

		int sensor,			/* what kind of sensor is connected */
			units,			/* the engineering units */
			wires,			/* how many wires are going to the sensor */
			linear,			/* what kind of linearization */
			cjc;			/* CJC off/on */

		float ptx,			/* the resistance ot 0deg C of a special platinum RTD */
			ssense,			/* the smart sense resistance in kOhm */
			urv, lrv, rng;	/* used for default urv/ lrv calculation */

		/* Setting the default values
		 */
		slen = 60;

		wires = 2;		/* DEFAULT: 2 Wires (2, 3, 4, Diff, SmartSense) */
		linear = 0;		/* DEFAULT: Temperature Linear (Temperature, Linear, Custom */
		cjc = 1;		/* DEFAULT: Don't Compensate (0 = Compensate) */
		units = 32;		/* DEFAULT: Celsius */
		ptx = 100.0;	/* DEFAULT: 100 Ohms */
		ssense = 100.0;	/* DEFAULT: 100 kOhms */

		/* take care of basics, we don't want the control system on line
		 * while we mess with the selecting the sensor!
		 */
		get_dictionary_string(remove_loop_from_auto, disp_string, slen); 
		ACKNOWLEDGE(disp_string);
		add_abort_method(return_to_normal);

		/* everything depends of the sensor: the number of wires; the 
		 * type of lenearization; whether CJC is used; and the ability
		 * to detect sensor breaks.
		 */
		do {
		}while (GET_DEV_VAR_VALUE("Select Sensor Type", 
			sensor_type) != BI_SUCCESS);

		switch (sensor = ivar_value (sensor_type) ) {
		case 1:			/* pt100 */
			selection = SELECT_FROM_LIST("Select Number of Wires",
				"Smart Sense;Differential;3 Wires;4 Wires");
			switch (selection){
			case 0:
				wires = 16;
				do {
				} while (GET_DEV_VAR_VALUE ("Enter Smart Sense Value",
					ssense_kOhm) != BI_SUCCESS);
				break;

			case 1:	wires = 5; break;
			case 3:	wires = 4; break;
			
			default: 
			case 2:	wires = 3; break;

			} /* end no. wires switch */

			break;

		case 2:			/* pt1000 */
		case 3:			/* d100 */
		case 4:			/* ptX */
		case 16:		/* ni100*/
		case 17:		/* ni1000*/
			if ( sensor == 4 ) {	/* ptX we need to know the r value */
				do {
				}while( GET_DEV_VAR_VALUE("Enter Pt X Resistance", 
					ptX_resistance) != BI_SUCCESS);

				ptx = ivar_value (ptX_resistance);

			}
				
			selection = SELECT_FROM_LIST("Select Number of Wires",
				"3 Wires;4 Wires");
			switch (selection){
			case 0:	wires = 3; break;
			case 1:	wires = 4; break;
			} /* end no. wires switch */

			break;

		case 32:									/* ohm */
		case 64:									/* mV */
			if(sensor == 32) {
				units = 37;							/* assign Ohms unit code */

				selection = SELECT_FROM_LIST("Select Number of Wires",
						"3 Wires;4 Wires");
				switch (selection){
				case 0:	wires = 3; break;
				case 1:	wires = 4; break;
				} /* end no. wires switch */
			}
			else {
				units = 36;							/* assign mV unit code */
				wires = 2;
			}

			selection = SELECT_FROM_LIST("Select Type of Linearization",
				"Linear with Input;Custom Linear");
			switch (selection){
			case 0:	linear = 1; break;
			case 1:	linear = 2; break;
			} /* end linearization switch */

			break;

		default:									/* All the T/C */

			if (sensor == 160) {					/* T/C X */
				units = 36;							/* assign mV unit code */
				linear = 2;							/* T/C X always custom */
			}

			if (sensor != 128) {
				selection = SELECT_FROM_LIST("Cold Junction Compensation",
					"On;Off");
				switch (selection){
				case 0:	cjc = 0; break;
				case 1:	cjc = 1; break;
				}	/* end cjc switch */
			}	/* end switch on CJC */

			selection = SELECT_FROM_LIST("Select Number of Wires",
				"Smart Sense;2 Wires");
			switch (selection){
			case 0:
				wires = 16;
				do {
				} while (GET_DEV_VAR_VALUE ("Enter Smart Sense Value",
					ssense_kOhm) != BI_SUCCESS);
				break;

			default: 
			case 1:	
				wires = 2; 
				break;
			}	/* end no. wires switch */

			break;		/* end default sensor type processing ( T/C ) */

		} /* end switch on sensor type */

		fassign( ptX_resistance , ptx);
		fassign( ssense_kOhm ,ssense);

		iassign( sensor_connection ,wires );
		iassign( sensor_linearization ,linear );
		iassign( sensor_CJC ,cjc );
		
		PUT_MESSAGE("Writing Sensor Configuration");

		do {
			SET_NUMBER_OF_RETRIES(0xff);
			IGNORE_ALL_COMM_STATUS();
			IGNORE_ALL_DEVICE_STATUS();
			IGNORE_ALL_RESPONSE_CODES();
			IGNORE_COMM_ERROR();
			IGNORE_NO_DEVICE();

			XMTR_IGNORE_ALL_COMM_STATUS();
			XMTR_IGNORE_ALL_DEVICE_STATUS();
			XMTR_IGNORE_ALL_RESPONSE_CODES();
			XMTR_IGNORE_COMM_ERROR();
			XMTR_IGNORE_NO_DEVICE();

			err = send_command(131); /*, status);		setting the device type etc */
			send_command(130);
		} while ( err == BI_NO_DEVICE && (
			sensor != ivar_value( sensor_type ) ||
			wires != ivar_value( sensor_connection ) ||
			linear != ivar_value( sensor_linearization ) ||
			cjc != ivar_value( sensor_CJC ) ) );

		if (status[STATUS_RESPONSE_CODE] != 0) {
			ACKNOWLEDGE ("PANIC: error sending command 131");
			display_response_status(131, status [STATUS_RESPONSE_CODE]);
		}

		RETRY_ON_NO_DEVICE();

		/* presetting the if the ranges.  we will read the USL and LSL
		 * after the sensor type is written and setting range to 90%
		 */
		send_command(14);		/* read the sensor limits */
		urv = fvar_value(temperature_USL);
		lrv = fvar_value(temperature_LSL);
		rng = (urv - lrv) / 2;
		lrv = lrv + rng * 0.1;
		urv = urv - rng * 0.1;
		fassign (temperature_LRV, lrv);
		fassign (temperature_URV, urv);
		iassign( input_units, units);

#if DEBUG
		ids[0]=VARID(temperature_URV);
		ids[1]=VARID(temperature_LRV);
		ids[2]=VARID(PV.RANGE_UNITS);
		acknowledge("URV = %{0},\nLRV= %{1},\nUnits = %{2}", ids);
#endif

		PUT_MESSAGE("Setting URV & LRV to Default Values");
		send(35, status);					/* setting the default ranges */
		if (status[STATUS_RESPONSE_CODE] != 0) {
			ACKNOWLEDGE ("PANIC: error sending command 35");
			display_response_status(35, status [STATUS_RESPONSE_CODE]);
		}
		send_command(15);

		if (linear == 2) {
			ACKNOWLEDGE("Remember to Enter/ Check Custom Linearization Values");
		}

		if ((cjc == 0) && (sensor == 160)) {
			ACKNOWLEDGE("Remember to Enter/ Check Custom CJC Values");
		}

		process_abort();		/* this will call return_to_normal */
	}
}


METHOD edit_err_offsets {
	LABEL "Edit Error Offsets";
	HELP "Set Error Offsets: This allows measurement errors to be compensated for by entering offests at the LRV and URV";
	CLASS CORRECTION;
	DEFINITION {
		char status[STATUS_SIZE];
		int choice;

		long ids[5];				/* used to hold VARIDS */

		long var_id[1];

		int use_abort_method;
		char disp_string1[60];
		char disp_string2[60];
		int slen1;
		int slen2;

		slen1=60;
		slen2=60;

		ids[ 0 ] = VARID(temperature_LRV);
		ids[ 1 ] = VARID(temperature_URV);
		ids[ 2 ] = VARID(sensor_error_units);
		ids[ 3 ] = VARID(sensor_error_LRV);
		ids[ 4 ] = VARID(sensor_error_URV);

		get_dictionary_string(remove_loop_from_auto,disp_string1,slen1);
		ACKNOWLEDGE(disp_string1);
		add_abort_method(return_to_normal);

		IGNORE_ALL_RESPONSE_CODES();
		IGNORE_ALL_DEVICE_STATUS();
		XMTR_IGNORE_ALL_DEVICE_STATUS();
		XMTR_IGNORE_ALL_RESPONSE_CODES();

		use_abort_method = 0;
		do {
			choice = select_from_list("Edit Error Offsets\nLRV %{3} %{2}\nURV %{4} %{2}}",
				ids,"Error @ LRV;Error @ URV;Done");

			switch ( choice ) {
			case 0:
				get_dev_var_value("Enter Error @LRV", ids, sensor_error_LRV);
				break;

			case 1:
				get_dev_var_value("Enter Error @URV", ids, sensor_error_URV);
				break;

			case 2:
			default:
				ACKNOWLEDGE("OK to send Error Offsets to the device?");
				PUT_MESSAGE("Writing Error Offsets...");

				send(135, status);
				if (status[STATUS_RESPONSE_CODE] != 0) {
					ACKNOWLEDGE ("Failed to write Error Offsets");
					display_response_status(135, status [STATUS_RESPONSE_CODE]);
					continue;
				}
				
				send(35, status);					/* setting the default ranges */
				if (status[STATUS_RESPONSE_CODE] != 0) {
					ACKNOWLEDGE ("PANIC: error sending command 35");
					display_response_status(35, status [STATUS_RESPONSE_CODE]);
				}
				process_abort();
				break;
			}
		} while ( choice < 3);
		process_abort();
	}
}
